import numpy as np
import cv2
from skimage import io
from picamera import PiCamera
from picamera.array import PiRGBArray
from time import sleep
from lib.caffe_detector import CaffeDetector
from lib.centroidtracker import CentroidTracker
from lib.trackableobject import TrackableObject
import json
import dlib

def main():
    f = open('config.json', 'r')
    config = json.loads(f.read())

    camera = PiCamera()
    camera.resolution = config['camera_resolution']
    camera.framerate = config['camera_framerate']
    rawCapture = PiRGBArray(camera, size=config['camera_resolution'])
    stream = camera.capture_continuous(rawCapture, format='bgr', use_video_port=True)

    # wait for camera to warm up
    sleep(0.1)

    detector = CaffeDetector(config['model_path'], 
                             config['prototxt_path'],
                             config['min_confidence'], 
                             config['classes'])

    cv2.namedWindow("Camera Detecting")

    centroid_tracker = CentroidTracker(maxDisappeared=config['max_disappeared'],
                                       maxDistance=config['max_distance'])
    trackers = []
    trackable_objects = {}

    frame_count = 0

    for f in stream:

        frame = f.array
        frame_count += 1

        # clear cached images
        rawCapture.truncate(0)
        frame = cv2.flip(frame, 0)
        
        # detection mode
        if frame_count % config['skip_frames'] == 0:

            boxes = detector.detect(frame, config['target_class_name'])
            
            trackers = []

            for box in boxes:
                tracker = dlib.correlation_tracker()
                rect = dlib.rectangle(*box)
                tracker.start_track(frame, rect)
                trackers.append(tracker)

        else: #tracking mode

            boxes = []
            
            for tracker in trackers:
                tracker.update(frame)
                pos = tracker.get_position()
                startX = int(pos.left())
                startY = int(pos.top())
                endX = int(pos.right())
                endY = int(pos.bottom())
                boxes.append((startX, startY, endX, endY))

        tracked_objects = centroid_tracker.update(boxes)

        for (objectID, centroid) in tracked_objects.items():

            trackable_object = trackable_objects.get(objectID, None)
            
            if trackable_object is not None:
                trackable_object.update_centroid(centroid)
            else:
                trackable_object = TrackableObject(objectID, centroid)
            
            trackable_objects[objectID] = trackable_object
            
            text = f'ID {objectID}'
            cv2.putText(frame, text, (centroid[0] - 10, centroid[1] - 10),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        0.5, (0, 255, 0), 2)
            cv2.circle(frame, (centroid[0], centroid[1]),
                        4, (0, 255, 0), -1)

        for box in boxes:
            startX, startY, endX, endY = box
            cv2.rectangle(frame, (startX, startY), (endX, endY), (255, 0, 0), 2)
            cv2.putText(frame, 
                        config['target_class_name'], 
                        (startX, startY), 
                        cv2.FONT_HERSHEY_SIMPLEX, 
                        1, (0, 255, 0), 3)

        cv2.imshow('Camera Detecting', frame)

        # Any key to stop video
        if cv2.waitKey(1) != -1:
            break


    stream.close()
    rawCapture.close()
    camera.close()
    cv2.destroyAllWindows()


if __name__ == "__main__":
    main()

