import numpy as np
import cv2
from skimage import io
from caffe_detector import CaffeDetector
from vino_detector import VinoDetector
from centroid_tracker import CentroidTracker
from trackable_object import TrackableObject
import math_util
import dlib
import time

class CVSensor:
    """ A computer vision based object counting sensor.
    """
    
    def __init__(self, stream, mediator, display_video, use_server, config):
        """
        Create a CVSensor object.

        stream: a stream of input images.
        mediator: an object which communicates to the desired system.
        display_video: True or False, should a window be displayed showing the detection and tracking?
        use_server: True or False, should the sensor communicate with the server using the supplied mediator?
        config: A configuration file dictionary. See config.toml for more details.
        """
        self.mediator = mediator
        self.config = config
        self.display_video = display_video
        self.use_server = use_server
        self.stream = stream

        w, h = stream.get_resolution()

        if display_video:
            # create a window for faster updates
            cv2.namedWindow("Camera Detecting")
            cv2.imshow('Camera Detecting', np.zeros((w, h, 3)))
            cv2.waitKey(1)

        if config['bin_path'].endswith('.bin'):
            self.detector = VinoDetector(config['bin_path'],
                                         config['desc_path'],
                                         config['min_confidence'],
                                         config['classes'])
        elif config['bin_path'].endswith('.caffemodel'):
            self.detector = CaffeDetector(config['model_path'], 
                                          config['prototxt_path'],
                                          config['min_confidence'], 
                                          config['classes'])
        else:
            print("Invalid model format. Should be '.caffemodel' or '.bin'")
            return
    
        # warm up the detector
        boxes = self.detector.detect(np.zeros((h, w, 3), dtype=np.uint8), self.config['target_class_name'], self.config['model_size'])


        # variable for keeping track of objects we are tracking
        self.centroid_tracker = CentroidTracker(maxDisappeared=config['max_disappeared'],
                                                maxDistance=config['max_distance'],
                                                framerate=config['camera_framerate'])
    
        # objects we are tracking. Associates ObjectID's to objects
        self.trackable_objects = {}

        self.in_count = 0

        # for a running average gamma
        self.avg_gamma = 1

        # change in time for a given frame
        self.delta_time = 0

        # for a running average framerate
        self.avg_fps = 30

        # correlation trackers
        self.trackers = []

        # build gamma tables
        pix_range=255

        test_start_time = time.time()
        # The following is used when pre processing is set up
        #self.gamma_to_index = { str(np.round(gamma,1)) : index for index, gamma in enumerate(np.arange(0.2, 1.6, 0.1))} 
        #self.gamma_table = np.array([[pix_range * (value / pix_range) ** gamma for value in range(256)] for gamma in np.arange(0.2, 1.6, 0.1)])


    def start(self):
        """
        Start the sensor detecting objects. Does not return anything.
        """

        self.stream.start()
        
        frame_count = 0

        for frame in self.stream:
            
            frame_start_time = time.time()

            #frame = self.pre_process(frame)
            frame_count += 1
            
            # detection mode
            if frame_count % self.config['skip_frames'] == 0:

                boxes = self.detector.detect(frame, self.config['target_class_name'], self.config['model_size'])
                self.init_trackers(boxes, frame)        

            else: #tracking mode

                boxes = self.update_trackers_and_create_boxes(frame)
            
            # associate boxes with existing tracked centroids
            tracked_centroids = self.centroid_tracker.update(boxes, self.delta_time)
            
            # updates trackable_objects or creates new ones based on new centroids  
            # also gets counts
            change_of_count = self.unify_objects_and_count(tracked_centroids)
       
            if change_of_count < 0:
                print('Car detected going out')
            elif change_of_count > 0:
                print('Car detected going in')

            self.in_count += change_of_count

            if change_of_count != 0 and self.use_server:
                self.mediator.change_count(change_of_count)


            frame_end_time = time.time()
            self.delta_time = frame_end_time - frame_start_time
            frame_rate = 1 / self.delta_time

            self.avg_fps = math_util.ema(self.avg_fps, frame_rate, math_util.num_samples_to_multiplier(60))

            if self.display_video:
                frame = frame.copy()
                self.decorate_frame(frame,
                           frame_rate=None,
                           boxes=boxes,
                           mark_objects=True,
                           tracked_centroids=tracked_centroids, 
                           draw_count=True,
                           draw_sep=True)

                cv2.imshow('Camera Detecting', frame)

                key = cv2.waitKey(1)
                
                # hitting return or escape will stop the sensor
                if key == ord('\r') or key == 27:
                    break
            
        self.stream.stop()

        if self.display_video:
            cv2.destroyAllWindows()


    def get_in_count(self):
        """
        Get the net number of cars that have been detected going into the location
        """
        return self.in_count


    def decorate_frame(self, frame, frame_rate=None, boxes=None, mark_objects=False, tracked_centroids=None, draw_count=False, draw_sep=False):
        """
        Add visual markings to the given frame
        frame: The image to draw over
        frame_rate: The fps of the sensor or None if no display is required.
        boxes: Any boxes to draw on the frame.
        mark_objects: True or False, should the objects be tagged?
        tracked_centroids: All the centroids to mark, or None if no centroids should be marked.
        draw_count: True or False, should the net count be drawn to the frame?
        draw_sep: True or False, should the dividing line (separater) be drawn on the frame?
        """
        if frame_rate is not None:
            cv2.putText(frame, f'FPS: {frame_rate:0.0f}', (10,25), cv2.FONT_HERSHEY_SIMPLEX, 0.4, (0, 255, 0), 2)

        if mark_objects:
            self.tag_objects(frame, tracked_centroids)
    
        if boxes is not None:
            self.draw_boxes(frame, boxes, self.config['target_class_name'])
    
        if draw_sep:
            cv2.line(frame, self.config['dividing_point_1'], 
                            self.config['dividing_point_2'], 
                            (200, 200, 0), 1)
    
        if draw_count:
            cv2.putText(frame, f'In Count: {self.in_count}', (10,10), cv2.FONT_HERSHEY_SIMPLEX, 0.4, (0, 255, 0), 2)


    def unify_objects_and_count(self, tracked_centroids):
        """
        Takes the new tracked centroids and unify's them with the existing trackable objects. It also uses the new tracked centroids to return a change in counts.
        """
        in_counts = 0

        for (objectID, centroid) in tracked_centroids.items():
            trackable_object = self.trackable_objects.get(objectID, None)

            if trackable_object is not None:
                trackable_object.update_centroid(centroid)
                in_counts += self.get_count(trackable_object)
            else:
                in_side = self.on_in_side(centroid)
                trackable_object = TrackableObject(objectID, 
                                               centroid, 
                                               in_side=in_side, 
                                               pos_recall=self.config['position_recall'],
                                               vector_recall=self.config['direction_recall'])
            self.trackable_objects[objectID] = trackable_object
        return in_counts


    def tag_objects(self, frame, tracked_centroids):
        """
        Draws the objectIDs on the given frame.
        """
        for (objectID, centroid) in tracked_centroids.items():
            to = self.trackable_objects.get(objectID, None)
            self.tag_object(frame, to, to.vector, objectID=objectID)


    def pre_process(self, frame):
        """
        Pre process the image to correct for lighting or color shifts.
        """
        pix_range = 255
        x_mean = np.mean(frame)
        
        if x_mean == 0:
            return frame

        gamma = np.log(127.5 / pix_range) / np.log(x_mean / pix_range)
        gamma = np.clip(gamma, 0.5, 1.5)

        # exponential running average
        multiplier = math_util.num_samples_to_multiplier(self.config['gamma_recall'])
        self.avg_gamma = math_util.ema(self.avg_gamma, gamma, multiplier)
        
        frame = cv2.LUT(frame, self.gamma_table[self.gamma_to_index[str(np.round(self.avg_gamma, 1))]])
        # frame = pix_range * (frame / pix_range) ** self.avg_gamma
        return frame.astype('uint8')


    def on_in_side(self, centroid):
        """
        Checks to see if a given centroid is inside the location (True) or outside the location (False).
        """
        dist = math_util.dist_to_line(centroid, self.config['dividing_point_1'], self.config['dividing_point_2']) 
        return dist > 0


    def get_count(self, trackable_object):
        """
        Using the given trackable_object, determines if that object should be counted in (+1) or out (-1) or neither (0).
        """
        line_dir = math_util.line_to_vector(self.config['dividing_point_1'], self.config['dividing_point_2'])
        rel_theta = math_util.relative_angle(trackable_object.vector, line_dir) 
        in_side =  self.on_in_side(trackable_object.avg_centroid) 

        # Check if the object has passed the line in either direction
        if in_side and not trackable_object.counted_in:
            trackable_object.counted_in = True
            trackable_object.counted_out = False
            return 1
        elif not in_side and not trackable_object.counted_out:
            trackable_object.counted_out = True
            trackable_object.counted_in = False
            return -1

        return 0


    def tag_object(self, frame, trackable_object, vector=None, objectID=None, drawCenter=True, drawAverage=True):
        """
        Marks the given object on the given frame.
        vector: A vector representing a line coming from the object center. Useful for direction of motion tags. Send None if no mark is desired.
        objectID: The ID to mark on the object or None if no mark is desired.
        drawCenter: True or False, should a center dot be drawn on the given object?
        drawAverage: True or False, should the average center dot (calculated over multiple frames) be drawn on the frame?
        """
        x,y = trackable_object.centroid
        avg_x, avg_y = trackable_object.avg_centroid

        if objectID is not None:
            text = f'ID {objectID}'
            cv2.putText(frame, text, (x - 10, y - 10),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        0.5, (0, 255, 0), 2)
        
        if drawCenter:
            cv2.circle(frame, (x, y),
                    4, (0, 255, 0), -1)

        if drawAverage:
            cv2.circle(frame, (int(avg_x),int(avg_y)), 4, (0, 255, 0), -1)

        if vector is not None and math_util.vector_can_divide(vector):
            # object direction indicator
            pt2 = (20 * vector / np.linalg.norm(vector)).astype(int) + trackable_object.centroid
            cv2.line(frame, trackable_object.centroid, pt2,  (200, 200, 0), 2)


    def draw_boxes(self, frame, boxes, label, font=cv2.FONT_HERSHEY_SIMPLEX, scale=.35, color=(0,255,0), thickness=1):
        """
        Draw the given boxes on the given frame with the given label
        """
        for box in boxes:
            
            startX, startY, endX, endY = box
            cv2.rectangle(frame, (startX, startY), (endX, endY), color, 2)
        
            if label is not None:
                cv2.putText(frame, 
                            label, 
                            (startX, startY), 
                            font, 
                            scale, 
                            color, 
                            thickness)
        return frame


    def init_trackers(self, boxes, frame):
        """
        Uses the detectors output (boxes) and the given frame to initialize trackers for use in the tracking phase of the sensor.
        """
        self.trackers = []
        
        for box in boxes:
            tracker = dlib.correlation_tracker()
            rect = dlib.rectangle(*box)
            tracker.start_track(frame, rect)
            self.trackers.append(tracker)


    def update_trackers_and_create_boxes(self, frame):
        """
        Use the given frame to update the object position in self.trackers
        """
        boxes = []
                
        for tracker in self.trackers:
            tracker.update(frame)
            pos = tracker.get_position()
            startX = int(pos.left())
            startY = int(pos.top())
            endX = int(pos.right())
            endY = int(pos.bottom())
            boxes.append((startX, startY, endX, endY))

        return boxes

