import numpy as np
from cv_sensor import CVSensor
from video_stream import VideoStream
from math_util import deg_to_perp_line
import pytest
import os
import re
import pytomlpp as toml

class DummyStream:
    def get_resolution(self):
        return (320, 240)

def setup_sensor(stream):

    f = open('config.toml', 'r')
    config = toml.loads(f.read())

    return CVSensor(stream=stream, 
                    mediator=None, 
                    display_video=False, 
                    use_server=False, 
                    config=config)


def parse_filename(filename):
    pat = re.compile(r"(?P<name>\w+)_(?P<fps>\d+)fps_(?P<deg>\d+)deg_(?P<sign>plus|minus)(?P<count>\d+)_(?P<cars>\d+)cars_\d+.mp4")
    matches = re.search(pat, filename)
    if not matches:
        raise Error("File name parsing error. Check that file is in the form NameOfFile_#fps_#deg_plus/minus#_#cars_#.mp4")
    name = matches.group('name')
    fps = int(matches.group('fps'))
    deg = int(matches.group('deg'))
    cars = int(matches.group('cars'))
    sign = -1 if matches.group('sign') == 'minus' else 1
    count = sign * int(matches.group('count'))
    return name, fps, deg, count, cars


def test_video_files_01():
    
    total_err = 0
    total_cars = 0
    log = []

    for filename in os.listdir('videos/'):
        if not filename.endswith('mp4'):
            continue

        name, fps, deg, true_count, cars = parse_filename(filename)
        stream = VideoStream('videos/' + filename, fps, loop=False)
        sensor = setup_sensor(stream)
        w, h = stream.get_resolution()

        center = np.array([w//2, h//2])
        pt1, pt2 = deg_to_perp_line(deg)
        pt1 = pt1 + center
        pt2 = -1 * pt2 + center
        sensor.config['dividing_point_1'] = pt1
        sensor.config['dividing_point_2'] = np.array([pt2[0], pt2[1]])
        sensor.start()
        count = sensor.get_in_count()

        if count != true_count:
            message = f'Count does not match in {filename}. Count = {count} while True Count = {true_count}.'
            print(message)
            log.append(message)

        total_err += np.abs(true_count - count)
        total_cars += cars

    rel_accuracy = total_err / total_cars
    print(f"Relative accuracy based on video files was {total_err}/{total_cars} = {rel_accuracy}%.")
    assert total_err / total_cars < 0.05, '\n'.join(log)


def test_on_in_side_01():
    sensor = setup_sensor(DummyStream())
    sensor.config['dividing_point_1'] = np.array([50,0])
    sensor.config['dividing_point_2'] = np.array([50, 200])
    centroid = np.array([50,50])
    assert sensor.on_in_side(centroid) == False
    centroid = np.array([150,50])
    assert sensor.on_in_side(centroid) == True

def test_on_in_side_02():
    sensor = setup_sensor(DummyStream())
    sensor.config['dividing_point_1'] = np.array([0,50])
    sensor.config['dividing_point_2'] = np.array([200, 50])
    centroid = np.array([100, 100])
    assert sensor.on_in_side(centroid) == False
    centroid = np.array([100, 0])
    assert sensor.on_in_side(centroid) == True

