import numpy as np

def deg_to_perp_line(deg, center=np.array([0,0])):
    """
    Returns two points that represent a line perpendicular to a given 
    angle in degrees and through a center point.
    """
    pt1 = center
    dir_vec = deg_to_vector(deg)
    perp_vec = rot_vector(dir_vec, np.pi / -2)
    pt2 = 100 * perp_vec + center
    return pt1, pt2


def deg_to_vector(deg):
    """
    Returns the unit vector in the direction of given angle in degrees.
    """
    return np.array([np.cos(np.deg2rad(deg)), np.sin(np.deg2rad(deg))])


def rot_vector(vector, rad):
    """
    Returns a new vector that is the same length as the given vector
    but rotated by the given radians.
    """
    rot = np.array([[np.cos(rad), -1 * np.sin(rad)],
                    [np.sin(rad),      np.cos(rad)]])
    return rot @ vector


def line_to_vector(line_pt1, line_pt2):
    """
    Returns a vector representing the direction a line is pointing (parallel).
    """
    line_dir = np.array([line_pt1[0] - line_pt2[0], 
                         line_pt1[1] - line_pt2[1]])
    return line_dir / np.linalg.norm(line_dir)


def vector_can_divide(vector):
    """
    Checks to make sure it's ok to divide by the given vector.
    """
    return np.NAN not in vector and 0 not in vector


def dist_to_line( point, line_pt1, line_pt2):
    """
    Returns the minimum distance from point to a line formed by two given points.
    Will return a negative distance depending on which side of the line it is on.
    """
    x_0, y_0 = point
    x_1, y_1 = line_pt1
    x_2, y_2 = line_pt2
    num = (x_2 - x_1) * (y_1 - y_0) - (x_1 - x_0) * (y_2 - y_1)
    denom = np.sqrt((x_2 - x_1)**2 + (y_2 - y_1)**2)
    return num / denom


def vector_to_angle(vector):
    """
    returns the angle formed by the given vector in radians
    """
    x,y = vector
    
    if x == 0 and y == 0:
        return 0 

    h = np.sqrt(x**2 + y**2)
    if x >= 0 and y >= 0:
        # quadrant 1
        return np.arcsin(y / h)
    if x < 0 and y >= 0:
        # quadrant 2
        return np.pi - np.arcsin(y / h)
    if x < 0 and y < 0:
        # quadrant 3
        return np.pi + np.arcsin(-1 * y / h)
    # quadrant 4
    return 2 * np.pi - np.arcsin(-1 * y / h)


def relative_angle(vector1, vector2):
    """
    returns the angle of the second vector relative to the first. 
    In the range [0,2pi)
    """
    theta = -1 * vector_to_angle(vector1)
    rotation_matrix = np.array([[np.cos(theta), -1 * np.sin(theta)],[np.sin(theta), np.cos(theta)]])
    adjusted_vector2 = rotation_matrix @ vector2
    return vector_to_angle(adjusted_vector2)


def ema(prev_avg, new_val, multiplier):
    """ Given the previous exponential moving average (EMA), the new value, and a multiplier, returns
    the new EMA.
    """
    return prev_avg + multiplier * (new_val - prev_avg)

def num_samples_to_multiplier(num_samples):
    """ Given the number of samples you'd like to average over, returns 
    the multiplier for use in EMA.
    """
    return 2 / (num_samples + 1)
