import numpy as np
import math_util

class TrackableObject:
    """
    A trackable object based on the one by pyimagesearch
    """
    def __init__(self, objectID, centroid, in_side=False, pos_recall=100, vector_recall=100):
        """
        Creates a new trackable object with no movement information
        
        Arguments:
        objectID - numeric ID for keeping track of objects
        centroid - the most recent location of the object as (x,y)
        vector_recall - OPTIONAL - the amount of past frames to consider when
                calculating a moving average vector.
        pos_recall - OPTIONAL - the amount of past frames to consider when
                calculating a moving average vector.

        """
        self.objectID = objectID
        self.centroid = centroid
        self.count = 1
        self.vector = np.array([0,0])
        self.avg_centroid = centroid
        self.counted_in = in_side
        self.counted_out = not in_side
        self.pos_mult = math_util.num_samples_to_multiplier(pos_recall)
        self.vect_mult = math_util.num_samples_to_multiplier(vector_recall)


    def update_centroid(self, centroid):
        """
        Updates the current location of the object. It also
        updates the direction vector via a moving exponential
        average.
        
        Arguments:
        centroid - a tuple containing the location as (x,y)
        
        """
        x,y = centroid
        old_x, old_y = self.centroid
        old_dx, old_dy = self.vector
        old_avg_x, old_avg_y = self.avg_centroid
        # running exponential average vector
        self.vector = np.array([math_util.ema(old_dx, (x-old_x), self.vect_mult),
                                math_util.ema(old_dy, (y-old_y), self.vect_mult)])
        if math_util.vector_can_divide(self.vector):
            self.vector = self.vector / np.linalg.norm(self.vector)
        # running exponential average position
        self.avg_centroid = np.array([math_util.ema(old_avg_x, x, self.pos_mult),
                                      math_util.ema(old_avg_y, y, self.pos_mult)])
        self.centroid = centroid
        self.count += 1

